#!/usr/bin/perl
use lib '/home/erealms/ethereal/mgmt/perl';

#################################################################################
# Created       : Martin Foster
# Modified      : 16-May-2004
#################################################################################
#
# Spelling - Daemon part of Ethereal Realms, designed to verify text and return
#            a list of improperly spelled words.
# Copyright (C) 2000-2006  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 29-26503 TWP RD 511
#                 Spruce Grove, Alberta
#                 T7Y 1G4
#
#################################################################################

use IO::Socket;							# SOcket handle
use IPC::Open2;							# Bidirectional reading
use POSIX;							# Set up socket server
use Symbol;							# Symbols

use Ethereal::Database;						# Database handle


#################################################################################
# Data members                                                                  #
#################################################################################

my $database;							# Database handle
my $kid;							# Server address
my $parent;							# Parent process

my %system;							# System table

#################################################################################
# Program Area                                                                  #
#################################################################################


	# Default
	$PREFORK		= 5;				# Children to maintain
	$MAX_CLIENTS_PER_CHILD  = 5;				# Each child can have				
	%children               = ();				# Record of children
	$children               = 0;				# Children count


	# Fork parent
	defined ($kid = fork) or die "Fork: $!\n";

	# Parent
	if ($kid)
	{

	}

	# Child does this
	else
	{
		# Be a gentleman
		# Complete detach
		POSIX::setsid()
		  or die("Session: $!");

		# Change to root directory
		chroot("/")
		  or warn("Could not chroot to /: $!");


		# Connect with database
		$database = new Ethereal::Database();

		# Connect and fetch
		$database->Connect();

		# Link with hash
		$database->GetHashSystem(\%system);


		# Create server
		$server = IO::Socket::INET->new
		(
			LocalPort => $system{'SetSpelPort'},
			Type      => SOCK_STREAM,
			Proto     => 'tcp',
			Reuse     => 1,
			Listen    => 10

		) or die("Making socket: $@\n");

		# Prefork
		for (1 .. $PREFORK)
		{
			# Spawn
			Childer($system{'SetSpelLoc'});
		}

		# Install signal handlers
		$SIG{CHLD} = \&REAPER;
		$SIG{INT}  = \&HUNTSMAN;


		# Maintain population
		while (1)
		{
			# Wait for signal
			sleep;

			# Child pool
			while ($children < $PREFORK) 
			{
				# Top up child pool
				Childer($system{'SetSpelLoc'});
			}
		}
	}

#################################################################################
# Sub-Routines
#################################################################################

####################
# Reaper
#
# This will look after the dead children.   Without as such, you will end
# things left behind.

sub REAPER
{
	####################
	# Data members
	my $pid;						# Process ID


	####################
	# Program area

	# Set sub-routine to child
	$SIG{CHLD} = \&REAPER;

	# Returned process
	$pid = wait;

	# Decrease amount of children
	$children--;

	# Delete process record
	delete $children{$pid};
}


####################
# Huntsman
#
# Handles a specific signal SIGINT

sub HUNTSMAN
{
	# Going to kill children
	local($SIG{CHLD}) = 'IGNORE';

	# Kill any recorded children
	kill 'INT' => keys(%children);

	# Clean up with dignity
	exit;
}


####################
# Childer
#
# Handles the process that is a child, at the same time
# will handle basic parts of the spell checking

sub Childer
{
	####################
	# Data members

	my $loc = shift;					# File location

	my $client;						# Client
	my $count;						# Counter
	my $pid;						# Process ID
	my $sigset;						# Setting of signal

	my $line;						# Captured line
	my $word;						# Word

	my $ipid;						# Internal process
	my $reader;						# Read filehandle
	my $writer;						# Write filehandle


	####################
	# Program area

	# Block signal for fork
	$sigset = POSIX::SigSet->new(SIGINT);
	sigprocmask(SIG_BLOCK, $sigset)
	  or die("Cannot block SIGINT for fork: $!\n");


	# Fork possible
	die "Fork: $!" unless defined ($pid = fork);


	# Parent
	if ($pid)
	{
		# Record the child birth and return
		sigprocmask(SIG_UNBLOCK, $sigset)
		  or die("Cannot unblock SIGINT for fork: $!\n");

		# Track child
		$children{$pid} = 1;

		# Increase child count
		$children++;

		# Retrun
		return;
	}


	# Child process
	# This child must die
	else
	{
		# Make SIGINT kill us as it did
		$SIG{INT} = 'DEFAULT';	

		# Unblock signals
		sigprocmask(SIG_UNBLOCK, $sigset)
		  or die("Cannot unblock SIGINT for fork: $!\n");


		# Handle connections until we have reached max
		for ($count=0; $count < $MAX_CLIENTS_PER_CHILD; $count++)
		{
			# Wait for use
			$client = $server->accept() or last;

			# Handling of the speller
			# Open bidirectrional pipe
			if ($ipid = open2($reader, $writer, $loc))
			{
				# Loop and verify
				while ($word = <$client>)
				{
					# Safety
					next unless (defined($word));
					next unless ($word ne '');

					# Client is done
					last unless ($word !~ /^0/);


					# Display word
					print $writer "$word\n";

					# Read pipe
					while ($line = <$reader>)
					{
						# Truncate
						chomp($line);

						# Determine use of line
						# Calculate and exit loop
						if ($line =~ /^([\#\&\*\+\-\?])/)
						{
							# Send complete output
							print $client "$line\n";

							# Final execution
							last;
						}	
					}
				}

				# Close filehandles
				close $reader;
				close $writer;

				# Wait for child
				waitpid($ipid, 0);
			}

			# Attempt to close
			close($client);
		}

		# Tidy up and finish
		exit;
	}
}
